'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _hat = require('hat');

var _hat2 = _interopRequireDefault(_hat);

var _omit = require('lodash/omit');

var _omit2 = _interopRequireDefault(_omit);

var _checkers = require('../utils/checkers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Debug.
 *
 * @type {Function}
 */

var debug = (0, _debug2.default)('nexus:connection');

/**
 * Connection.
 *
 * @type {Connection}
 */

var Connection = function (_Emitter) {
  _inherits(Connection, _Emitter);

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  function Connection() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, Connection);

    var _this = _possibleConstructorReturn(this, (Connection.__proto__ || Object.getPrototypeOf(Connection)).call(this));

    var server = config.server,
        transport = config.transport,
        context = config.context;

    _this.server = server;
    _this.transport = transport;
    _this.context = context;
    _this.id = (0, _hat2.default)();
    _this.subscriptions = {};

    transport.on('readable', function () {
      var data = _this.transport.read();
      if (data === null) return _this.close();
      _this.onMessage(data);
    });

    transport.on('error', function (error) {
      _this.emit('error', error);
    });
    return _this;
  }

  /**
   * Close the connection.
   */

  _createClass(Connection, [{
    key: 'close',
    value: function close() {
      var subscriptions = this.subscriptions;


      for (var key in subscriptions) {
        this.unsubscribe(key);
      }

      this.emit('close');
    }

    /**
     * Unsubscribe from a document by `id`.
     *
     * @param {String} id
     */

  }, {
    key: 'unsubscribe',
    value: function unsubscribe(id) {
      var subscriptions = this.subscriptions;

      var sub = subscriptions[id];
      if (!sub) return;

      sub.push(null);
      delete subscriptions[id];
    }

    /**
     * Send a `message` to the client.
     *
     * @param {Object} message
     */

  }, {
    key: 'send',
    value: function send(message) {
      var id = this.id;

      debug('send', { id: id, message: message });
      this.transport.write(message);
    }

    /**
     * Send a connect message to the client.
     */

  }, {
    key: 'sendConnect',
    value: function sendConnect() {
      this.send({
        action: 'connect',
        id: this.id
      });
    }

    /**
     * Send a `snapshot` message to the client for a `document` and `action`.
     *
     * @param {String} document
     * @param {String} action
     * @param {Object} snapshot
     */

  }, {
    key: 'sendSnapshot',
    value: function sendSnapshot(document, action, snapshot) {
      var message = {
        action: action,
        document: document
      };

      if (snapshot) {
        message.snapshot = snapshot;
      }

      this.send(message);
    }

    /**
     * Send an `operation` message to the client for a `document` and `action`.
     *
     * @param {String} document
     * @param {String} action
     * @param {Object} operation
     */

  }, {
    key: 'sendOperation',
    value: function sendOperation(document, action, operation) {
      var message = {
        action: action,
        document: document,
        id: operation.id,
        version: operation.version
      };

      if (operation.data !== undefined) {
        message.data = operation.data;
      }

      this.send(message);
    }

    /**
     * Send `operations` to the server for a `document`.
     *
     * @param {String} document
     * @param {String} action
     * @param {Array} operations
     */

  }, {
    key: 'sendOperations',
    value: function sendOperations(document, action, operations) {
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = operations[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var op = _step.value;

          this.sendOperation(document, action, op);
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    }

    /**
     * Send an unsubscribe message to the client for a `document`.
     *
     * @param {String} document
     */

  }, {
    key: 'sendUnsubscribe',
    value: function sendUnsubscribe(document) {
      this.send({
        action: 'unsubscribe',
        document: document
      });
    }

    /**
     * On message.
     *
     * @param {Object} message
     * @return {Mixed}
     */

  }, {
    key: 'onMessage',
    value: function () {
      var _ref = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee(message) {
        var id, server;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                id = this.id, server = this.server;


                debug('onMessage', { id: id, message: message });

                (0, _checkers.checkMessage)(message);

                _context.prev = 3;
                _context.t0 = message.action;
                _context.next = _context.t0 === 'connect' ? 7 : _context.t0 === 'fetch' ? 10 : _context.t0 === 'submit' ? 13 : _context.t0 === 'subscribe' ? 16 : _context.t0 === 'unsubscribe' ? 19 : 22;
                break;

              case 7:
                _context.next = 9;
                return this.onMessageConnect(message);

              case 9:
                return _context.abrupt('return', _context.sent);

              case 10:
                _context.next = 12;
                return this.onMessageFetch(message);

              case 12:
                return _context.abrupt('return', _context.sent);

              case 13:
                _context.next = 15;
                return this.onMessageSubmit(message);

              case 15:
                return _context.abrupt('return', _context.sent);

              case 16:
                _context.next = 18;
                return this.onMessageSubscribe(message);

              case 18:
                return _context.abrupt('return', _context.sent);

              case 19:
                _context.next = 21;
                return this.onMessageUnsubscribe(message);

              case 21:
                return _context.abrupt('return', _context.sent);

              case 22:
                _context.next = 29;
                break;

              case 24:
                _context.prev = 24;
                _context.t1 = _context['catch'](3);

                this.emit('error', _context.t1);
                server.emit('error', _context.t1);

                this.send(_extends({}, message, {
                  error: {
                    code: _context.t1.code,
                    message: _context.t1.message,
                    data: _context.t1.data
                  }
                }));

              case 29:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this, [[3, 24]]);
      }));

      function onMessage(_x2) {
        return _ref.apply(this, arguments);
      }

      return onMessage;
    }()

    /**
     * On "connect" message.
     *
     * @param {Object} message
     * @return {Object}
     */

  }, {
    key: 'onMessageConnect',
    value: function () {
      var _ref2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2(message) {
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                this.sendConnect();

              case 1:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function onMessageConnect(_x3) {
        return _ref2.apply(this, arguments);
      }

      return onMessageConnect;
    }()

    /**
     * On "fetch" message.
     *
     * @param {Object} message
     * @return {Object|Void}
     */

  }, {
    key: 'onMessageFetch',
    value: function () {
      var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3(message) {
        var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

        var _options$action, action, context, server, document, version, _document$split, _document$split2, id, ops, snapshot;

        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _options$action = options.action, action = _options$action === undefined ? 'fetch' : _options$action;
                context = this.context, server = this.server;
                document = message.document, version = message.version;
                _document$split = document.split(':'), _document$split2 = _slicedToArray(_document$split, 1), id = _document$split2[0];
                ops = [];
                snapshot = void 0;

                if (!(version == null)) {
                  _context3.next = 12;
                  break;
                }

                _context3.next = 9;
                return server.findSnapshot(id, { context: context });

              case 9:
                snapshot = _context3.sent;
                _context3.next = 15;
                break;

              case 12:
                _context3.next = 14;
                return server.listOperations(id, version, undefined, { context: context });

              case 14:
                ops = _context3.sent;

              case 15:

                this.sendOperations(document, 'broadcast', ops);
                this.sendSnapshot(document, action, snapshot);

              case 17:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function onMessageFetch(_x4) {
        return _ref3.apply(this, arguments);
      }

      return onMessageFetch;
    }()

    /**
     * On "submit" message.
     *
     * @param {Object} message
     * @return {Object}
     */

  }, {
    key: 'onMessageSubmit',
    value: function () {
      var _ref4 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4(message) {
        var context, server, document, _document$split3, _document$split4, id, ops, ack;

        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                context = this.context, server = this.server;
                document = message.document;
                _document$split3 = document.split(':'), _document$split4 = _slicedToArray(_document$split3, 1), id = _document$split4[0];
                ops = void 0;
                _context4.prev = 4;
                _context4.next = 7;
                return server.submit(id, message, { context: context });

              case 7:
                ops = _context4.sent;
                _context4.next = 14;
                break;

              case 10:
                _context4.prev = 10;
                _context4.t0 = _context4['catch'](4);

                if (!(_context4.t0.code !== 'operation_duplicate')) {
                  _context4.next = 14;
                  break;
                }

                throw _context4.t0;

              case 14:

                if (ops) {
                  this.sendOperations(document, 'broadcast', ops);
                }

                ack = (0, _omit2.default)(message, 'data');

                this.sendOperation(document, 'submit', ack);

              case 17:
              case 'end':
                return _context4.stop();
            }
          }
        }, _callee4, this, [[4, 10]]);
      }));

      function onMessageSubmit(_x6) {
        return _ref4.apply(this, arguments);
      }

      return onMessageSubmit;
    }()

    /**
     * On "subscribe" message.
     *
     * @param {Object} message
     * @return {Object|Void}
     */

  }, {
    key: 'onMessageSubscribe',
    value: function () {
      var _ref5 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6(message) {
        var _this2 = this;

        var context, server, subscriptions, id, document, version, _document$split5, _document$split6, documentId, subscription, previous;

        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                context = this.context, server = this.server, subscriptions = this.subscriptions, id = this.id;
                document = message.document, version = message.version;
                _document$split5 = document.split(':'), _document$split6 = _slicedToArray(_document$split5, 1), documentId = _document$split6[0];
                _context6.next = 5;
                return server.subscribe(documentId, version, { context: context });

              case 5:
                subscription = _context6.sent;
                previous = subscriptions[documentId];


                if (previous) {
                  previous.push(null);
                }

                subscriptions[documentId] = subscription;
                subscription.on('readable', _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5() {
                  var op;
                  return regeneratorRuntime.wrap(function _callee5$(_context5) {
                    while (1) {
                      switch (_context5.prev = _context5.next) {
                        case 0:
                          op = subscription.read();

                          if (!(op === null)) {
                            _context5.next = 5;
                            break;
                          }

                          delete subscriptions[document];
                          subscription.emit('close');
                          return _context5.abrupt('return');

                        case 5:
                          if (!op.id.startsWith(id + ':')) {
                            _context5.next = 7;
                            break;
                          }

                          return _context5.abrupt('return');

                        case 7:

                          // Otherwise, broadcast the op to the client.
                          _this2.sendOperation(document, 'broadcast', op);

                        case 8:
                        case 'end':
                          return _context5.stop();
                      }
                    }
                  }, _callee5, _this2);
                })));

                // Call into the fetch handler to fetch the current snapshot or operations.
                _context6.next = 12;
                return this.onMessageFetch(message, { action: 'subscribe' });

              case 12:
              case 'end':
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function onMessageSubscribe(_x7) {
        return _ref5.apply(this, arguments);
      }

      return onMessageSubscribe;
    }()

    /**
     * On "unsubscribe" message.
     *
     * @param {Object} message
     */

  }, {
    key: 'onMessageUnsubscribe',
    value: function () {
      var _ref7 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7(message) {
        var document, _document$split7, _document$split8, id;

        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                document = message.document;
                _document$split7 = document.split(':'), _document$split8 = _slicedToArray(_document$split7, 1), id = _document$split8[0];

                this.unsubscribe(id);
                this.sendUnsubscribe(document);

              case 4:
              case 'end':
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function onMessageUnsubscribe(_x8) {
        return _ref7.apply(this, arguments);
      }

      return onMessageUnsubscribe;
    }()
  }]);

  return Connection;
}(_events2.default);

/**
 * Export.
 *
 * @type {Connection}
 */

exports.default = Connection;